/*=============================================================================
    Copyright (c) 2014 Christoph Weiss

    Distributed under the Boost Software License, Version 1.0. (See accompanying
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef FUSION_INCLUDE_HASH
#define FUSION_INCLUDE_HASH

#include <boost/fusion/sequence/hash.hpp>

#endif
