// Copyright (c) 2006, 2007 Julio M. Merino Vidal
// Copyright (c) 2008 Ilya Sokolov, Boris Schaeling
// Copyright (c) 2009 Boris Schaeling
// Copyright (c) 2010 Felipe Tanus, Boris Schaeling
// Copyright (c) 2011 Jeff Flinn, Boris Schaeling
// Copyright (c) 2012 Boris Schaeling
// Copyright (c) 2016 Klemens D. Morgenstern
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_PROCESS_HPP
#define BOOST_PROCESS_HPP

#include <boost/process/bind_launcher.hpp>
#include <boost/process/default_launcher.hpp>
#include <boost/process/environment.hpp>
#include <boost/process/error.hpp>
#include <boost/process/execute.hpp>
#include <boost/process/exit_code.hpp>
#include <boost/process/ext.hpp>
#include <boost/process/pid.hpp>
#include <boost/process/popen.hpp>
#include <boost/process/process.hpp>
#include <boost/process/shell.hpp>
#include <boost/process/start_dir.hpp>
#include <boost/process/stdio.hpp>

#endif
