//
// Copyright (c) 2024 Julien Blanc (julien.blanc@tgcm.eu)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// Official repository: https://github.com/boostorg/json
//

#ifndef BOOST_JSON_DEBUG_PRINTERS_HPP

#include <boost/json/detail/gdb_printers.hpp>

#endif
