#ifndef BOOST_UUID_UUID_HASH_HPP_INCLUDED
#define BOOST_UUID_UUID_HASH_HPP_INCLUDED

// Copyright (c) 2018 James E. King III
// Distributed under the Boost Software License, Version 1.0.
// https://www.boost.org/LICENSE_1_0.txt

// This header is no longer needed and is only retained
// for backward compatibility.

#include <boost/uuid/uuid.hpp>

#endif // #ifndef BOOST_UUID_UUID_HASH_HPP_INCLUDED
