#ifndef BOOST_LAMBDA2_HPP_INCLUDED
#define BOOST_LAMBDA2_HPP_INCLUDED

// Copyright 2020 Peter Dimov
// Distributed under the Boost Software License, Version 1.0.
// https://www.boost.org/LICENSE_1_0.txt

#include <boost/lambda2/lambda2.hpp>

#endif // #ifndef BOOST_LAMBDA2_HPP_INCLUDED
