//
// Copyright (c) 2024 Mohammad Nejati
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// Official repository: https://github.com/boostorg/beast
//

#ifndef BOOST_BEAST_HTTP_BASIC_FILE_BODY_FWD_HPP
#define BOOST_BEAST_HTTP_BASIC_FILE_BODY_FWD_HPP

namespace boost {
namespace beast {
namespace http {

template<class File>
struct basic_file_body;

} // http
} // beast
} // boost

#endif
