//
// Copyright (c) 2024 Klemens Morgenstern (klemens.morgenstern@gmx.net)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_COBALT_IO_HPP
#define BOOST_COBALT_IO_HPP

#include <boost/cobalt/io/acceptor.hpp>
#include <boost/cobalt/io/buffer.hpp>
#include <boost/cobalt/io/datagram_socket.hpp>
#include <boost/cobalt/io/endpoint.hpp>
#include <boost/cobalt/io/file.hpp>
#include <boost/cobalt/io/ops.hpp>
#include <boost/cobalt/io/pipe.hpp>
#include <boost/cobalt/io/random_access_device.hpp>
#include <boost/cobalt/io/random_access_file.hpp>
#include <boost/cobalt/io/read.hpp>
#include <boost/cobalt/io/resolver.hpp>
#include <boost/cobalt/io/seq_packet_socket.hpp>
#include <boost/cobalt/io/serial_port.hpp>
#include <boost/cobalt/io/signal_set.hpp>
#include <boost/cobalt/io/sleep.hpp>
#include <boost/cobalt/io/socket.hpp>
#include <boost/cobalt/io/steady_timer.hpp>
#include <boost/cobalt/io/stream.hpp>
#include <boost/cobalt/io/stream_file.hpp>
#include <boost/cobalt/io/stream_socket.hpp>
#include <boost/cobalt/io/system_timer.hpp>
#include <boost/cobalt/io/write.hpp>


#endif //BOOST_COBALT_IO_HPP
