//  Copyright (c) 2001-2011 Hartmut Kaiser
//
//  Distributed under the Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(BOOST_SPIRIT_KARMA_FORMAT_DEC_01_2009_0716AM)
#define BOOST_SPIRIT_KARMA_FORMAT_DEC_01_2009_0716AM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/karma/stream/format_manip.hpp>
#include <boost/spirit/home/karma/stream/format_manip_attr.hpp>

#endif
